from .email import EmailNotifier
from .telegram import TelegramNotifier


class Notifier:
    """
    A notifier factory which chooses the notifier based on the "type" param.
    """

    NOTIFIERS = [EmailNotifier, TelegramNotifier]

    def __new__(cls, type: str, *args, **kwargs):
        for notifier in cls.NOTIFIERS:
            if type == notifier.TYPE:
                return notifier(*args, **kwargs)
        raise ValueError(f"Unknown notifier type: {type}")



import subprocess
from tinydb import TinyDB, Query

from .config import DB_FILENAME


class Check:
    """
    A command which is run on a given host with a certain expected output.
    """

    def __init__(self, host, username, command, expected, description=""):
        self.host = host
        self.username = username
        self.command = command
        self.expected = expected
        self.description = description

    def _execute_ssh_command(self) -> str:
        """Executes an SSH command and returns stdout and stderr."""
        result = subprocess.run(
            [
                "ssh",
                "-l",
                self.username,
                self.host,
                "--",
                self.command,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        def _decode(b):
            return b.decode("utf-8").strip() if b else ""

        return _decode(result.stdout), _decode(result.stderr)

    @property
    def _last_stdout(self) -> str:
        db = TinyDB(DB_FILENAME)
        Stdout = Query()

        values = db.search(Stdout.id == self._db_id)
        return values[-1]["stdout"] if values else ""

    @_last_stdout.setter
    def _last_stdout(self, value) -> None:
        db = TinyDB(DB_FILENAME)
        Stdout = Query()

        if self._last_stdout:
            db.update({"id": self._db_id, "stdout": value}, Stdout.id == self._db_id)
        else:
            db.insert({"id": self._db_id, "stdout": value})

    @property
    def _db_id(self) -> str:
        return hash(f"{self.host}:{self.command}")

    def run(self) -> tuple[bool, bool, str, str]:
        """Executes the command and checks the results."""

        stdout, stderr = self._execute_ssh_command()

        stdout_changed = stdout != self._last_stdout
        self._last_stdout = stdout

        return stdout_changed, stdout == self.expected, stdout, stderr



import logging


__all__ = ["TelegramNotifier"]


class TelegramNotifier:
    """
    A telegram notifier.
    """

    TYPE = "telegram"
    ENCODING = "UTF-8"

    def __init__(self, token: str, chat_id: str):
        # Only import the telegram dependency when it's needed
        try:
            from telegram import Bot
        except ModuleNotFoundError:
            logging.error("Please install the `python-telegram-bot` package.")
            exit(1)

        self._bot = Bot(token=token)
        self._chat_id = chat_id

    def notify(self, text: str) -> None:
        """Sends a message."""
        self._bot.sendMessage(chat_id=self._chat_id, text=text)



import os

from .check import Check
from .config import CONFIG
from .notifiers import Notifier


__all__ = ["WatchDog"]


class WatchDog:
    """
    Executes commands over SSH and notifies when the output is unexpected.

    Will keep track of the state to avoid excessive notification spamming.
    """

    def __init__(self):
        self._ssh_config: dict[str, str] = CONFIG.get("ssh", {})

        self._checks: list[Check] = [
            Check(h, self._ssh_config.get("username", "fikkie"), **c)
            for h, cs in CONFIG.get("servers", {}).items()
            for c in cs
        ]

        self._notifiers: list[Notifier] = [
            Notifier(**n) for n in CONFIG.get("notifiers", [])
        ]

    def notify(self, msg: str, icon: str = "") -> None:
        """Sends a notification using all available notifiers."""
        for notifier in self._notifiers:
            notifier.notify(
                f"{icon} {msg}" if notifier.ENCODING == "UTF-8" and icon else msg
            )

    def tick(self) -> None:
        """Perform checks."""
        for check in self._checks:
            stdout_changed, stdout_expected, stdout, stderr = check.run()

            if stdout_changed:
                if stdout_expected:
                    self.notify(f"{check.host}: {check.description} OK", "🟢")
                else:
                    self.notify(
                        f"{check.host}: {check.description} NOT OK\n{stdout} (stderr: {stderr})",
                        "🔴",
                    )

    def heartbeat(self) -> None:
        """Send a notification to let user know the watchdog's alive."""
        self.notify("❤️ I'm still awake.")



from celery import Celery
from celery.schedules import crontab

import os

from .config import BROKER_DIR
from .watchdog import WatchDog


app = Celery(__name__)
app.conf.update(
    {
        "broker_url": "filesystem://",
        "broker_transport_options": {
            "data_folder_in": os.path.join(BROKER_DIR, "out"),
            "data_folder_out": os.path.join(BROKER_DIR, "out"),
            "data_folder_processed": os.path.join(BROKER_DIR, "processed"),
        },
        "result_persistent": False,
        "task_serializer": "json",
        "result_serializer": "json",
        "accept_content": ["json"],
    }
)
app.conf.beat_schedule = {
    "tick": {"task": "fikkie.main.tick", "schedule": 60},
    "heartbeat": {
        "task": "fikkie.main.heartbeat",
        "schedule": crontab(hour=12, minute=0),  # UTC 12:00
    },
}


@app.task
def tick():
    watchdog = WatchDog()
    watchdog.tick()


@app.task
def heartbeat():
    watchdog = WatchDog()
    watchdog.heartbeat()



import logging
import os
import yaml


__all__ = ["BASE_DIR", "CONFIG_FILE", "BROKER_DIR", "CONFIG", "DB_FILENAME"]

BASE_DIR = os.getenv("FIKKIE_BASE_DIR", os.path.expanduser("~/.fikkie"))
CONFIG_FILE = os.getenv("FIKKIE_CONFIG", os.path.join(BASE_DIR, "config.yaml"))
BROKER_DIR = os.getenv("FIKKIE_BROKER_DIR", os.path.join(BASE_DIR, "broker"))
DB_FILENAME = os.getenv("FIKKIE_DB_FILENAME", os.path.join(BASE_DIR, "db.json"))


def load_config(filename: str) -> dict:
    """Parse the config from a given file."""
    try:
        with open(filename, "r") as f:
            return yaml.safe_load(f) or {}
    except FileNotFoundError as e:
        logging.warning("Please run `fikkie --init`.")
        exit(1)
    except yaml.YAMLError as e:
        logging.error(f"Could not parse config: {e}")
        exit(1)


CONFIG = load_config(CONFIG_FILE)



import logging
import smtplib
import ssl


__all__ = ["EmailNotifier"]


class EmailNotifier:
    """
    An e-mail notifier.
    """

    TYPE = "email"
    ENCODING = "ASCII"

    def __init__(
        self,
        recipient: str,
        email: str,
        password: str,
        smtp_server: str,
        smtp_port: int = 465,
    ):
        self._recipient = recipient
        self._email = email
        self._password = password
        self._smtp_server = smtp_server
        self._smtp_port = smtp_port

    def notify(self, text: str) -> None:
        """Sends a message."""
        message = f"""Subject: {text}

        Fikkie wants to notify you about the following message:

        {text}
        """

        context = ssl.create_default_context()

        with smtplib.SMTP_SSL(
            self._smtp_server, self._smtp_port, context=context
        ) as server:
            server.login(self._email, self._password)
            server.sendmail(self._email, self._recipient, message)



from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="fikkie",
    version="0.2.1",
    description="Lightweight monitoring tool",
    license="MIT",
    long_description=long_description,
    author="Joris Hartog (@nootr)",
    long_description_content_type="text/markdown",
    url="https://nootr.github.io/fikkie/",
    packages=["fikkie", "fikkie.notifiers"],
    scripts=["scripts/fikkie"],
    install_requires=[
        "celery==5.2.1",
        "PyYAML==6.0",
        "tinydb==4.5.2",
    ],
)
